from __future__ import annotations

import pickle
from typing import Any, ClassVar, Generic, NoReturn, Type, TypeVar

from glassjar.constants import DB_NAME
from glassjar.db import DB, DatabaseManager, create_table
from glassjar.exceptions import DoesNotExist
from glassjar.field import Field

T = TypeVar("T", bound="Model")


class QuerySet(Generic[T]):
    def __init__(self, objs: list[T]) -> None:
        self.__objs = objs
        self.__index = 0

    def __iter__(self) -> "QuerySet":
        return self

    def __next__(self) -> T | NoReturn:
        try:
            obj = self.__objs[self.__index]
        except IndexError:
            raise StopIteration

        self.__index += 1
        return obj

    def __repr__(self) -> str:
        return f"<QuerySet {tuple(obj for obj in self.__objs)} >"

    def __len__(self) -> int:
        return len(self.__objs)

    def as_dict(self) -> dict[int, dict[str, Any]]:
        return {obj.id: obj.as_dict() for obj in self.__objs}

    def count(self) -> int:
        return len(self.__objs)

    def first(self) -> "QuerySet" | T:
        try:
            return self.__objs[0]
        except IndexError:
            return QuerySet([])

    def last(self) -> "QuerySet" | T:
        try:
            return self.__objs[-1]
        except IndexError:
            return QuerySet([])


class QueryManager:
    def __init__(self, name: str) -> None:
        self.table_name = f"{name}_table"

    def get(self, id: int) -> T | NoReturn:
        with DB(DB_NAME, write_back=True) as db:
            try:
                obj = db.db["tables"][self.table_name]["records"][id]
                return pickle.loads(obj)
            except KeyError:
                raise DoesNotExist("Object does not exist.")

    def all(self) -> QuerySet:
        with DB(DB_NAME, write_back=True) as db:
            try:
                values = db.db["tables"][self.table_name]["records"].values()
                objs = [pickle.loads(val) for val in values]
                return QuerySet(objs)
            except KeyError:
                return QuerySet([])

    def count(self) -> int:
        return self.all().count()

    def first(self) -> QuerySet | T:
        return self.all().first()

    def last(self) -> QuerySet | T:
        return self.all().last()


class BaseModel(type):
    def __new__(
        mcs, cls_name: str, bases: tuple, cls_dict: Any
    ) -> Type[T] | "BaseModel":
        slots = []
        fields = cls_dict.get("__annotations__", {})
        fields.update({"id": int})

        for field_name, field_type in fields.items():
            field = Field(field_name, field_type)
            cls_dict[field_name] = field
            slots.append(field.storage_name)

        cls_dict["__slots__"] = slots
        obj = super().__new__(mcs, cls_name, bases, cls_dict)
        setattr(obj, "records", QueryManager(cls_name))
        setattr(obj, "table_name", f"{cls_name}_table")
        create_table(f"{cls_name}_table")

        return obj


class Model(DatabaseManager, metaclass=BaseModel):
    records: ClassVar[QueryManager]

    def __init__(self, **fields: Any) -> None:
        super().__init__(**fields)

    def __eq__(self, other):
        if self.as_dict == self.as_dict:
            return True
        else:
            return False

    def __repr__(self) -> str:
        kwargs = ", ".join(
            f"{key}={getattr(self, key)!r}" for key in self.fields
        )
        return f"{type(self).__name__}({kwargs})"

    def as_dict(self) -> dict[str, Any]:
        return {key: getattr(self, key) for key in self.fields}

    def update(self) -> None:
        self._update_record()

    def delete(self, id: int) -> None:
        self._delete_record(id)

    def save(self) -> "Model":
        if hasattr(self, "id"):
            self._update_record()
        else:
            self.__create_record()
        return self

    def __create_record(self) -> "Model":
        with DB(DB_NAME, write_back=True) as db:
            table = db.db["tables"][self.table_name]
            setattr(self, "id", table["index"])
            table["records"].update({table["index"]: pickle.dumps(self)})
            table["index"] += 1
        return self



from typing import Any, Callable


class Field:
    def __init__(self, name: str, field_type: Callable) -> None:
        self.name = name
        self.storage_name = f"_{name}"
        self.field_type = field_type

    def __get__(self, instance: Any, owner: Any = None) -> Any:
        if instance is None:
            return self
        return getattr(instance, self.storage_name)

    def __set__(self, instance: Any, value: Any) -> None:
        if self.field_type != type(value):
            raise TypeError(
                f"Types are incompatible. Expected type: {self.field_type} | Given type: {type(value)}"
            )
        setattr(instance, self.storage_name, value)



import pickle
from typing import Any

from glassjar.constants import DB_NAME
from glassjar.db import DB
from glassjar.exceptions import DoesNotExist


class QueryManager:
    def __init__(self, name: str) -> None:
        self.table_name = f"{name}_table"

    def get(self, id: int) -> Any:
        with DB(DB_NAME, write_back=True) as db:
            try:
                obj = db.db["tables"][self.table_name]["records"][id]
                return pickle.loads(obj)
            except KeyError:
                raise DoesNotExist("Object does not exist.")

    def all(self) -> list:
        with DB(DB_NAME, write_back=True) as db:
            try:
                values = db.db["tables"][self.table_name]["records"].values()
                objs = [pickle.loads(val) for val in values]
                return objs
            except KeyError:
                return []

    def count(self) -> int:
        return len(self.all())

    def first(self) -> Any:
        try:
            return self.all()[0]
        except IndexError:
            return []

    def last(self) -> Any:
        try:
            return self.all()[-1]
        except IndexError:
            return []



from __future__ import annotations

import contextlib
import os
import pickle
from io import BytesIO
from pickle import Pickler, Unpickler
from types import TracebackType
from typing import Any, ClassVar, Hashable

from glassjar.constants import DB_NAME
from glassjar.exceptions import DoesNotExist


class DB:
    def __init__(self, file_name: str, write_back: bool = False) -> None:
        self.file_name = file_name
        self.write_back = write_back
        self.cache: dict[Hashable, Any] = {}
        self.db: dict[Hashable, Any] = {}
        self.create_or_set_db()

    def __getitem__(self, key: Hashable) -> Any:
        try:
            value = self.cache[key]
        except KeyError:
            value = Unpickler(BytesIO(self.db[key])).load()
            if self.write_back:
                self.cache[key] = value
        return value

    def __setitem__(self, key: Hashable, value: Any) -> None:
        if self.write_back:
            self.cache[key] = value
        f = BytesIO()
        Pickler(f, pickle.HIGHEST_PROTOCOL).dump(value)
        self.db[key] = f.getvalue()

    def __delitem__(self, key: Hashable) -> None:
        del self.db[key]
        with contextlib.suppress(KeyError):
            del self.cache[key]

    def __enter__(self) -> "DB":
        return self

    def __exit__(
        self,
        exc_type: type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        self.close()

    def create_or_set_db(self) -> None:
        try:
            if os.path.getsize(self.file_name):
                with open(self.file_name, "rb") as fp:
                    self.db = pickle.load(fp)
        except FileNotFoundError:
            self.db["tables"] = {}
            with open(self.file_name, "wb") as fp:
                fp.write(b"")

    def get(self, key: Hashable, default: Any = None) -> Any:
        if key in self.db:
            return self.db[key]
        return default

    def close(self) -> None:
        if self.write_back and self.cache:
            for key, entry in self.cache.items():
                self[key] = entry
            self.cache = {}
        with open(self.file_name, "wb") as fp:
            pickle.dump(self.db, fp)

    def initialize_table(self, table_name: str) -> None:
        if self.db["tables"].get(table_name) is None:
            self.db["tables"][table_name] = {"index": 1, "records": {}}


class DatabaseManager:
    __slots__ = "fields"
    table_name: ClassVar[str]
    id: ClassVar[int]

    def __init__(self, **fields: dict[str, Any]) -> None:
        self.fields = fields
        for field_name, field_value in self.fields.items():
            setattr(self, field_name, field_value)

    def _get_record(self, id: int) -> bytes:
        with DB(DB_NAME, write_back=True) as db:
            try:
                obj = db.db["tables"][self.table_name]["records"][id]
                return obj
            except KeyError:
                raise DoesNotExist("Object does not exist.")

    def _set_record(self, id: int, value: Any) -> None:
        with DB(DB_NAME, write_back=True) as db:
            value = pickle.dumps(value)
            db.db["tables"][self.table_name]["records"][id] = value

    def _update_record(self) -> None:
        db_obj = pickle.loads(self._get_record(self.id))

        for field_name, field_value in self.fields.items():
            obj_value = getattr(self, field_name)
            if getattr(db_obj, field_name) != obj_value:
                setattr(db_obj, field_name, obj_value)

        self._set_record(self.id, db_obj)

    def _delete_record(self, id: int) -> None:
        with DB(DB_NAME, write_back=True) as db:
            try:
                del db.db["tables"][self.table_name]["records"][id]
            except KeyError:
                raise DoesNotExist("Object does not exist.")


def create_table(table_name: str) -> None:
    with DB(DB_NAME, write_back=True) as db:
        db.initialize_table(table_name)



from pathlib import Path

from setuptools import setup

CURRENT_DIR = Path(__file__).parent


def get_long_description():
    readme_md = CURRENT_DIR / "README.md"
    with open(readme_md, encoding="utf8") as ld_file:
        return ld_file.read()


setup(
    name="glassjar",
    version="0.1.0",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    description="Glassjar is the pickled database that provide Object-Relational Mapper.",
    keywords=["pickled database", "orm"],
    author="Furkan Onder",
    author_email="furkanonder@protonmail.com",
    url="https://github.com/furkanonder/glassjar/",
    license="MIT",
    python_requires=">=3.8",
    packages=["glassjar"],
    install_requires=[],
    extras_require={},
    zip_safe=False,
    include_package_data=False,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        "Operating System :: OS Independent ",
    ],
)
